@@ -1,9 +1,62 @@
 use {
-    crate::{Chunk, FunctionFlags, Op},
+    crate::{Chunk, Op},
     koto_parser::ConstantIndex,
     std::{convert::TryInto, fmt, sync::Arc},
 };
 
+#[derive(Debug)]
+#[repr(u8)]
+pub enum TypeId {
+    List,
+    Tuple,
+}
+
+impl TypeId {
+    fn from_byte(byte: u8) -> Result<Self, u8> {
+        if byte == Self::List as u8 {
+            Ok(Self::List)
+        } else if byte == Self::Tuple as u8 {
+            Ok(Self::Tuple)
+        } else {
+            Err(byte)
+        }
+    }
+}
+
+pub struct FunctionFlags {
+    pub instance_function: bool,
+    pub variadic: bool,
+    pub generator: bool,
+}
+
+impl FunctionFlags {
+    pub const INSTANCE: u8 = 0b0000001;
+    pub const VARIADIC: u8 = 0b0000010;
+    pub const GENERATOR: u8 = 0b0000100;
+
+    pub fn from_byte(byte: u8) -> Self {
+        Self {
+            instance_function: byte & Self::INSTANCE == Self::INSTANCE,
+            variadic: byte & Self::VARIADIC == Self::VARIADIC,
+            generator: byte & Self::GENERATOR == Self::GENERATOR,
+        }
+    }
+
+    pub fn as_byte(&self) -> u8 {
+        let mut result = 0;
+        if self.instance_function {
+            result |= Self::INSTANCE;
+        }
+        if self.variadic {
+            result |= Self::VARIADIC;
+        }
+        if self.generator {
+            result |= Self::GENERATOR;
+        }
+        result
+    }
+}
+
 /// Decoded instructions produced by an [InstructionReader] for execution in the runtime
 pub enum Instruction {
     Error {
@@ -13,10 +66,6 @@ pub enum Instruction {
         target: u8,
         source: u8,
     },
-    DeepCopy {
-        target: u8,
-        source: u8,
-    },
     SetEmpty {
         register: u8,
     },
@@ -300,6 +349,14 @@ pub enum Instruction {
         register: u8,
         constant: ConstantIndex,
     },
+    CheckType {
+        register: u8,
+        type_id: TypeId,
+    },
+    CheckSize {
+        register: u8,
+        size: usize,
+    },
 }
 
 impl fmt::Display for Instruction {
@@ -308,7 +365,6 @@ impl fmt::Display for Instruction {
         match self {
             Error { message } => unreachable!(message),
             Copy { .. } => write!(f, "Copy"),
-            DeepCopy { .. } => write!(f, "DeepCopy"),
             SetEmpty { .. } => write!(f, "SetEmpty"),
             SetBool { .. } => write!(f, "SetBool"),
             SetNumber { .. } => write!(f, "SetNumber"),
@@ -372,6 +428,8 @@ impl fmt::Display for Instruction {
             TryStart { .. } => write!(f, "TryStart"),
             TryEnd => write!(f, "TryEnd"),
             Debug { .. } => write!(f, "Debug"),
+            CheckType { .. } => write!(f, "CheckType"),
+            CheckSize { .. } => write!(f, "CheckSize"),
         }
     }
 }
@@ -382,9 +440,6 @@ impl fmt::Debug for Instruction {
         match self {
             Error { message } => unreachable!(message),
             Copy { target, source } => write!(f, "Copy\t\tresult: {}\tsource: {}", target, source),
-            DeepCopy { target, source } => {
-                write!(f, "DeepCopy\tresult: {}\tsource: {}", target, source)
-            }
             SetEmpty { register } => write!(f, "SetEmpty\tresult: {}", register),
             SetBool { register, value } => {
                 write!(f, "SetBool\t\tresult: {}\tvalue: {}", register, value)
@@ -738,6 +793,12 @@ impl fmt::Debug for Instruction {
             Debug { register, constant } => {
                 write!(f, "Debug\t\tregister: {}\tconstant: {}", register, constant)
             }
+            CheckType { register, type_id } => {
+                write!(f, "CheckType\tregister: {}\ttype: {:?}", register, type_id)
+            }
+            CheckSize { register, size } => {
+                write!(f, "CheckSize\tregister: {}\tsize: {}", register, size)
+            }
         }
     }
 }
@@ -822,10 +883,6 @@ impl Iterator for InstructionReader {
                 target: get_byte!(),
                 source: get_byte!(),
             }),
-            Op::DeepCopy => Some(DeepCopy {
-                target: get_byte!(),
-                source: get_byte!(),
-            }),
             Op::SetEmpty => Some(SetEmpty {
                 register: get_byte!(),
             }),
@@ -1172,6 +1229,19 @@ impl Iterator for InstructionReader {
                 register: get_byte!(),
                 constant: get_u32!() as ConstantIndex,
             }),
+            Op::CheckType => {
+                let register = get_byte!();
+                match TypeId::from_byte(get_byte!()) {
+                    Ok(type_id) => Some(CheckType { register, type_id }),
+                    Err(byte) => Some(Error {
+                        message: format!("Unexpected value for CheckType id: {}", byte),
+                    }),
+                }
+            }
+            Op::CheckSize => Some(CheckSize {
+                register: get_byte!(),
+                size: get_byte!() as usize,
+            }),
             _ => Some(Error {
                 message: format!("Unexpected opcode {:?} found at instruction {}", op, op_ip),
             }),
